// -------------------------------------------------------------------
//  Copyright (c) 2011-2013 TIBCO Software, Inc.
//  All rights reserved.
//  For more information, please contact:
//  TIBCO Software Inc., Palo Alto, California, USA
//	Oracle sample code by Stefano Ghio - sghio@tibco.com
// -------------------------------------------------------------------
package com.tibco.shared_all.persistence;

import com.tibco.as.space.DateTime;
import com.tibco.as.space.FieldDef;
import com.tibco.as.space.SpaceDef;
import com.tibco.as.space.Tuple;
import com.tibco.as.space.persistence.PutOp;
import com.tibco.as.space.persistence.TakeOp;

import java.sql.*;
import java.util.*;

public class ASJdbcConnection
{

    protected Connection connection;
    protected String tableName;
    protected Collection<FieldDef> nonKeyFields;
    protected Collection<FieldDef> keyFields;
    protected Collection<FieldDef> allFields;
    protected PreparedStatement deleteCall, insertCall, updateCall, selectCall;
    protected Map<FieldDef.FieldType, String> mapping;

    public ASJdbcConnection (String driverCls, String connectionURL, Properties properties) throws SQLException, ClassNotFoundException
    {
        Class.forName(driverCls);
        connection = DriverManager.getConnection(connectionURL, properties);
    }

    public ASJdbcConnection (Connection connection, SpaceDef spaceDef) throws SQLException
    {
        this.connection = connection;
    }

    public void init (SpaceDef spaceDef) throws SQLException
    {
        mapping = getMapping();
        tableName = spaceDef.getName();
        allFields = spaceDef.getFieldDefs();
        keyFields = new ArrayList<FieldDef>();
        nonKeyFields = new ArrayList<FieldDef>();
        Collection<String> fieldNames = spaceDef.getKeyDef().getFieldNames();

        for (FieldDef field : allFields)
        {
            String fieldName = field.getName();
            if (fieldNames.contains(fieldName))
            {
                keyFields.add(field);
            }
            else
            {
                nonKeyFields.add(field);
            }
        }

        deleteCall = connection.prepareStatement(createDeleteQuery());
        insertCall = connection.prepareStatement(createInsertQuery());
        updateCall = connection.prepareStatement(createUpdateQuery());
        selectCall = connection.prepareStatement(createSelectQuery());

        // try to define table def, it is already defined -- ignore
        try
        {
            String createQuery = getCreateTableQuery();
            Statement statement = connection.createStatement();
            statement.execute(createQuery);
            statement.close();
        }
        catch (SQLException ex)
        {
            // throw new RuntimeException(ex);
        }
    }

    public void close ()
    {
        try
        {
            insertCall.close();
            deleteCall.close();
            updateCall.close();
            selectCall.close();
            connection.close();
        }
        catch (Exception ex)
        {
            throw new RuntimeException(ex);
        }
    }

    public Tuple loadTuple (Tuple keyTuple)
    {
        try
        {
            selectCall.clearParameters();

            int index = 1;
            for (FieldDef field : keyFields)
            {
                setField(index, field, keyTuple, selectCall);
                index++;
            }

            ResultSet result = selectCall.executeQuery();
            if (result.next())
            {
                Tuple tuple = Tuple.create();
                for (FieldDef field : allFields)
                {
                    readField(result, field, tuple);
                }
                return tuple;
            }
        }
        catch (SQLException ex)
        {
            throw new RuntimeException(ex);
        }
        return null;
    }

    public Collection<Tuple> load (int maxSize, int offset)
    {
        Collection<Tuple> collection = new ArrayList<Tuple>();
        try
        {
            Statement statement = connection.createStatement();
            ResultSet result = statement.executeQuery("SELECT * FROM " + tableName);
            while (result.next())
            {
                Tuple tuple = Tuple.create();
                for (FieldDef field : allFields)
                {
                    readField(result, field, tuple);
                }
                collection.add(tuple);
            }

            result.close();
            statement.close();
        }
        catch (Exception ex)
        {
            throw new RuntimeException(ex);
        }

        return collection;
    }

    public void process (PutOp putOp)
    {
        try
        {
            int index = 1;
            Tuple tuple = putOp.getTuple();
            PreparedStatement ps = null;
            String op = "";
            if (putOp.hasOldTuple())    // Update query
            {
                op = "update";
                ps = updateCall;
                updateCall.clearParameters();
                for (FieldDef field : nonKeyFields)
                {
                    setField(index, field, tuple, updateCall);
                    index++;
                }
                for (FieldDef field : keyFields)
                {
                    setField(index, field, tuple, updateCall);
                    index++;
                }
            }
            else                          // insert query
            {
                op = "put";
                ps = insertCall;
                insertCall.clearParameters();
                for (FieldDef field : allFields)
                {
                    setField(index, field, tuple, insertCall);
                    index++;
                }
            }
            ps.execute();

            if (ps.getUpdateCount() == 1)
            {
                System.out.println(op + " was successful");
            }
            else
            {
                System.out.println(op + " was not successful");
            }
        }
        catch (Exception ex)
        {
            throw new RuntimeException(ex);
        }
    }

    public void process (TakeOp takeOp)
    {
        try
        {
            deleteCall.clearParameters();
            int index = 1;
            Tuple tuple = takeOp.getTuple();

            for (FieldDef field : keyFields)
            {
                setField(index, field, tuple, deleteCall);
                index++;
            }

            deleteCall.execute();

            if (deleteCall.getUpdateCount() == 1)
            {
                System.out.println("Take was successful");
            }
            else
            {
                System.out.println("Take was not successful");
            }
        }
        catch (Exception ex)
        {
            throw new RuntimeException(ex);
        }
    }

    public void setSQLMapping (FieldDef.FieldType fieldType, String SQLType)
    {
        mapping.put(fieldType, SQLType);
    }

    public String getSQLMapping (FieldDef.FieldType fieldType)
    {
        return mapping.get(fieldType);
    }

    // -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    // SQL queries to create table, insert/update/delete a row from table
    // -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
    private void readField (ResultSet rs, FieldDef field, Tuple tuple) throws SQLException
    {
        String fieldName = field.getName();

        switch (field.getType())
        {
            case BLOB:
                byte[] bytes = rs.getBytes(fieldName);
                if (bytes != null)
                {
                    tuple.putBlob(fieldName, bytes);
                }
                break;
            case DATETIME:
                Timestamp ts = rs.getTimestamp(fieldName);
                if (ts != null)
                {
                    tuple.putDateTime(fieldName, DateTime.create(ts.getTime()));
                }
                break;
            case BOOLEAN:
                tuple.putBoolean(fieldName, rs.getBoolean(fieldName));
                break;
            case CHAR:
                tuple.putChar(fieldName, (Character) rs.getObject(fieldName));
                break;
            case DOUBLE:
                tuple.putDouble(fieldName, rs.getDouble(fieldName));
                break;
            case FLOAT:
                tuple.putFloat(fieldName, rs.getFloat(fieldName));
                break;
            case INTEGER:
                tuple.putInt(fieldName, rs.getInt(fieldName));
                break;
            case LONG:
                tuple.putLong(fieldName, rs.getLong(fieldName));
                break;
            case SHORT:
                tuple.putShort(fieldName, rs.getShort(fieldName));
                break;
            case STRING:
                tuple.putString(fieldName, rs.getString(fieldName));
                break;
        }
    }

    private void setField (int index, FieldDef field, Tuple tuple, PreparedStatement ps) throws SQLException
    {
        Object value = tuple.get(field.getName());
        if (value == null)
        {
            ps.setObject(index, null);
            return;
        }

        switch (field.getType())
        {
            case BLOB:
                byte[] bytes = (byte[]) value;
                ps.setBytes(index, bytes);
                break;
            case DATETIME:
                DateTime datetime = (DateTime) value;
                Timestamp ts = new Timestamp(datetime.getTimeInMillis());
                ps.setTimestamp(index, ts);
                break;
            case BOOLEAN:
            case CHAR:
            case DOUBLE:
            case FLOAT:
            case INTEGER:
            	ps.setObject(index, value);
            	break;
            case LONG:
            case SHORT:
            case STRING:
                ps.setObject(index, value);
                break;
        }
    }

    protected String getCreateTableQuery ()
    {
        String query = "";
        query += "CREATE TABLE " + tableName + " (";

        for (FieldDef field : allFields)
        {
            query += getFieldName(field) + " " + mapping.get(field.getType());
            if (!field.isNullable())
            {
                query += " not";
            }
            query += " null, ";
        }

        query += "Primary Key (";

        Iterator<FieldDef> iterator = keyFields.iterator();
        while (iterator.hasNext())
        {
            FieldDef field = iterator.next();
            query += getFieldName(field);
            if (iterator.hasNext())
            {
                query += ",";
            }
        }

        query += ")";   // close primary key constraint
        query += ")";   // close table def

        return query;
    }

    private String createDeleteQuery ()
    {
        String query = "";
        query += "DELETE FROM " + tableName + " WHERE ";

        for (FieldDef field : keyFields)
        {
            query += getFieldName(field) + " = ?, ";
        }

        return query.substring(0, query.length() - 2);
    }

    private String createSelectQuery ()
    {
        String query = "";
        query += "SELECT * FROM " + tableName + " WHERE ";

        for (FieldDef field : keyFields)
        {
            query += getFieldName(field) + " = ?, ";
        }

        return query.substring(0, query.length() - 2);
    }

    private String createInsertQuery ()
    {
        String query = "";
        query += "INSERT INTO " + tableName + " (";

        Iterator<FieldDef> iterator = allFields.iterator();
        while (iterator.hasNext())
        {
            FieldDef field = iterator.next();
            query += getFieldName(field);
            if (iterator.hasNext())
            {
                query += ", ";
            }
        }

        query += ") VALUES(" + prepareQs(allFields.size()) + ")";
        return query;
    }

    private String createUpdateQuery ()
    {
        String query = "";
        query += "Update " + tableName;

        query += " set ";
        Iterator<FieldDef> iterator = nonKeyFields.iterator();
        while (iterator.hasNext())
        {
            FieldDef field = iterator.next();
            query += getFieldName(field) + " = ?";
            if (iterator.hasNext())
            {
                query += ", ";
            }
        }

        query += " where ";
        iterator = keyFields.iterator();
        while (iterator.hasNext())
        {
            FieldDef field = iterator.next();
            query += getFieldName(field) + " = ?";

            if (iterator.hasNext())
            {
                query += " and ";
            }
        }

        return query;
    }

    private String prepareQs (int size)
    {
        String result = "";
        for (int i = 0; i < size; i++)
        {
            result += "?";
            if (i + 1 != size)
            {
                result += ",";
            }
        }

        return result;
    }

    protected String getFieldName (FieldDef field)
    {
        return field.getName();
    }

    protected Map<FieldDef.FieldType, String> getMapping ()
    {
        Map<FieldDef.FieldType, String> map = new HashMap<FieldDef.FieldType, String>();
        map.put(FieldDef.FieldType.BLOB, "BYTEA");
        map.put(FieldDef.FieldType.BOOLEAN, "BIT");
        map.put(FieldDef.FieldType.CHAR, "CHAR");
        map.put(FieldDef.FieldType.DATETIME, "TIMESTAMP");
        map.put(FieldDef.FieldType.DOUBLE, "DOUBLE PRECISION");
        map.put(FieldDef.FieldType.FLOAT, "FLOAT");
        map.put(FieldDef.FieldType.INTEGER, "INTEGER");
        map.put(FieldDef.FieldType.LONG, "REAL");
        map.put(FieldDef.FieldType.SHORT, "SMALLINT");
        map.put(FieldDef.FieldType.STRING, "CHARACTER VARYING");
        return map;
    }
}
